#!/usr/bin/env python
#    "$Name:  $";
#    "$Header:  $";
#=============================================================================
#
# file :        Experiment_Details.py
#
# description : Python source for the Experiment_Details and its commands.
#                The class is derived from Device. It represents the
#                CORBA servant object which will be accessed from the
#                network. All commands which can be executed on the
#                Experiment_Details are implemented in this file.
#
# project :     TANGO Device Server
#
# $Author:  $
#
# $Revision:  $
#
# $Log:  $
#
# copyleft :    European Synchrotron Radiation Facility
#               BP 220, Grenoble 38043
#               FRANCE
#
#=============================================================================
#          This file is generated by POGO
#    (Program Obviously used to Generate tango Object)
#
#         (c) - Software Engineering Group - ESRF
#=============================================================================
#


import PyTango
import sys
import time

#==================================================================
#   Experiment_Details Class Description:
#
#         Holds useful information about current experiment
#
#==================================================================
#     Device States Description:
#
#   DevState.ON :
#   DevState.FAULT :
#==================================================================


class Experiment_Details(PyTango.Device_4Impl):

#--------- Add you global variables here --------------------------

#------------------------------------------------------------------
#    Device constructor
#------------------------------------------------------------------
    def __init__(self,cl, name):
        PyTango.Device_4Impl.__init__(self,cl,name)
        Experiment_Details.init_device(self)

#------------------------------------------------------------------
#    Device destructor
#------------------------------------------------------------------
    def delete_device(self):
        print ("[Device delete_device method] for device",self.get_name())

#------------------------------------------------------------------
#    Device initialization
#------------------------------------------------------------------
    def init_device(self):
        #print ("In ", self.get_name(), "::init_device()")
        self.UserLogin = ""
        self.UserId = ""
        self.ActiveInvestigation = ""
        self.StaffMember = False
        self.Logged_in = False
        self.DataPath = ""
        self.StorageMngrDev = None
        self.ProposalList = []
        self.InvestigationList = []
        self.ActionLog = []
        self.free_space_readout_time = 0
        self.attr_FreeSpaceScratch_read = 0
        self.attr_FreeSpaceOnline_read = 0
        self.attr_FreeSpaceScratch_TB_read = 0
        self.attr_FreeSpaceOnline_TB_read = 0
        self.last_time_activity = time.time()
        #
        self.set_state(PyTango.DevState.ON)
        self.get_device_properties(self.get_device_class())
        # Manually manage some memorized attribute values
        self.db = PyTango.Database()
        mem_attr_list = ['Logged_in','StaffMember','UserId','ActiveInvestigation']
        dev_attr_props = self.db.get_device_attribute_property(self.get_name(),mem_attr_list)
        attr_prop = dev_attr_props['Logged_in']
        if '__value' in attr_prop.keys():
            self.Logged_in = bool(attr_prop['__value'][0] == "True")
        attr_prop = dev_attr_props['StaffMember']
        if '__value' in attr_prop.keys():
            self.StaffMember = bool(attr_prop['__value'][0] == "True")
        attr_prop = dev_attr_props['ActiveInvestigation']
        if '__value' in attr_prop.keys():
            self.ActiveInvestigation = attr_prop['__value'][0]
        attr_prop = dev_attr_props['UserId']
        if '__value' in attr_prop.keys():
            self.UserId = attr_prop['__value'][0]
        #
        if self.Logged_in:
            if len(self.ActiveInvestigation.strip()):
                self.set_active_investigation()
            else:
                self.retrieve_proposal_investigations()

#------------------------------------------------------------------
#    Always excuted hook method
#------------------------------------------------------------------
    def always_executed_hook(self):
        #print ("In ", self.get_name(), "::always_excuted_hook()")
        if self.StorageMngrDev == None:
            self.connect_to_storage_manager()
        # Logout if there are no activities for 2 days
        if (time.time() - self.last_time_activity) > (2 * 86400):
            self.Logout()
            self.last_time_activity = time.time()

#==================================================================
#
#    Experiment_Details read/write attribute methods
#
#==================================================================
#------------------------------------------------------------------
#    Read Attribute Hardware
#------------------------------------------------------------------
    def read_attr_hardware(self,data):
        #print ("In ", self.get_name(), "::read_attr_hardware()")
        pass

#------------------------------------------------------------------
#    Read UserLogin attribute
#------------------------------------------------------------------
    def read_UserLogin(self, attr):
        #print ("In ", self.get_name(), "::read_UserLogin()")

        #    Add your own code here
        attr_User_read = self.UserLogin
        attr.set_value(attr_User_read)


#------------------------------------------------------------------
#    Write UserLogin attribute
#------------------------------------------------------------------
    def write_UserLogin(self, attr):
        #print ("In ", self.get_name(), "::write_UserLogin()")
        self.UserLogin = attr.get_write_value()

#------------------------------------------------------------------
#    Read ActiveInvestigation attribute
#------------------------------------------------------------------
    def read_ActiveInvestigation(self, attr):
        #print ("In ", self.get_name(), "::read_ActiveInvestigation()")

        #    Add your own code here
        attr_ActiveInvestigation_read = self.ActiveInvestigation
        attr.set_value(attr_ActiveInvestigation_read)


#------------------------------------------------------------------
#    Write ActiveInvestigation attribute
#------------------------------------------------------------------
    def write_ActiveInvestigation(self, attr):
        #print ("In ", self.get_name(), "::write_ActiveInvestigation()")

        #    Add your own code here
        #if not self.Logged_in:
        #  raise Exception("User not logged in!")
        self.ActiveInvestigation =  attr.get_write_value()
        #
        self.set_active_investigation()
        self.memorize_status()


#------------------------------------------------------------------
#    Read Logged_in attribute
#------------------------------------------------------------------
    def read_Logged_in(self, attr):
        #print ("In ", self.get_name(), "::read_Logged_in()")

        #    Add your own code here
        attr_Logged_in_read = self.Logged_in
        attr.set_value(attr_Logged_in_read)

#------------------------------------------------------------------
#    Read UserId attribute
#------------------------------------------------------------------
    def read_UserId(self, attr):
        #print ("In ", self.get_name(), "::read_UserId()")

        #    Add your own code here

        attr_UserId_read = self.UserId
        attr.set_value(attr_UserId_read)


#------------------------------------------------------------------
#    Read StaffMember attribute
#------------------------------------------------------------------
    def read_StaffMember(self, attr):
        #print ("In ", self.get_name(), "::read_StaffMember()")

        #    Add your own code here
        attr_StaffMember_read = self.StaffMember
        attr.set_value(attr_StaffMember_read)


#------------------------------------------------------------------
#    Read DataPath attribute
#------------------------------------------------------------------
    def read_DataPath(self, attr):
        #print ("In ", self.get_name(), "::read_DataPath()")

        #    Add your own code here
        attr_DataPath_read = self.DataPath
        attr.set_value(attr_DataPath_read)


#------------------------------------------------------------------
#    Write DataPath attribute
#------------------------------------------------------------------
    def write_DataPath(self, attr):
        #print ("In ", self.get_name(), "::write_DataPath()")
        self.DataPath = attr.get_write_value()

#------------------------------------------------------------------
#    Read ProposalList attribute
#------------------------------------------------------------------
    def read_ProposalList(self, attr):
        #print ("In ", self.get_name(), "::read_ProposalList()")

        #    Add your own code here

        attr_ProposalList_read = self.ProposalList
        attr.set_value(attr_ProposalList_read)

#------------------------------------------------------------------
#    Read InvestigationList attribute
#------------------------------------------------------------------
    def read_InvestigationList(self, attr):
        #print ("In ", self.get_name(), "::read_InvestigationList()")

        #    Add your own code here

        attr_InvestigationList_read = self.InvestigationList
        attr.set_value(attr_InvestigationList_read)

#------------------------------------------------------------------
#    Read FreeSpaceScratch attribute
#------------------------------------------------------------------
    def read_FreeSpaceScratch(self, attr):
        #print ("In ", self.get_name(), "::read_FreeSpaceScratch()")

        #    Add your own code here
        self.update_disk_usage_info()
        attr.set_value(self.attr_FreeSpaceScratch_read)


#------------------------------------------------------------------
#    Read FreeSpaceOnline attribute
#------------------------------------------------------------------
    def read_FreeSpaceOnline(self, attr):
        #print ("In ", self.get_name(), "::read_FreeSpaceOnline()")

        #    Add your own code here
        self.update_disk_usage_info()
        attr.set_value(self.attr_FreeSpaceOnline_read)

#------------------------------------------------------------------
#    Read FreeSpaceScratch_TB attribute
#------------------------------------------------------------------
    def read_FreeSpaceScratch_TB(self, attr):
        #print ("In ", self.get_name(), "::read_FreeSpaceScratch_TB()")

        #    Add your own code here
        self.update_disk_usage_info()
        attr.set_value(self.attr_FreeSpaceScratch_TB_read)


#------------------------------------------------------------------
#    Read FreeSpaceOnline_TB attribute
#------------------------------------------------------------------
    def read_FreeSpaceOnline_TB(self, attr):
        #print ("In ", self.get_name(), "::read_FreeSpaceOnline_TB()")

        #    Add your own code here
        self.update_disk_usage_info()
        attr.set_value(self.attr_FreeSpaceOnline_TB_read)

#------------------------------------------------------------------
#    Read ActionLog attribute
#------------------------------------------------------------------
    def read_ActionLog(self, attr):
        #print ("In ", self.get_name(), "::read_ActionLog()")

        #    Add your own code here

        attr_ActionLog_read = self.ActionLog
        attr.set_value(attr_ActionLog_read)

#==================================================================
#
#    Experiment_Details command methods
#
#==================================================================
#------------------------------------------------------------------
#    Login()
#------------------------------------------------------------------
    def Login(self, password):
        self.StaffMember = False
        self.Logged_in = False
        only_resp = self.CheckCredentials([self.UserLogin, password, 'nofull'])
        self.UserId = only_resp.split(":")[0]
        if only_resp.find(":ADMIN") > 0:
            self.StaffMember = True
            self.log_action("Authenticated user is staff member")
        self.Logged_in = True
        self.retrieve_proposal_investigations()
        self.memorize_status()

#------------------------------------------------------------------
#    Logout()
#------------------------------------------------------------------
    def Logout(self):
        self.Logged_in = False
        self.UserId = ""
        self.ActiveInvestigation = ""
        self.ProposalList = []
        self.InvestigationList = []
        self.StaffMember = False
        self.log_action("Authenticated user logout")
        self.memorize_status()

#------------------------------------------------------------------
#    Login()
#------------------------------------------------------------------
    def CheckCredentials(self, argin):
        if self.StorageMngrDev == None:
            self.connect_to_storage_manager()
        UserLogin = argin[0]
        password = argin[1]
        arg_in = [UserLogin,password,self.TAG]
        self.log_action("CheckCredentials user %s" % UserLogin)
        try:
            resp = self.StorageMngrDev.Authenticate(arg_in)
        except Exception as e:
            print(e)
        self.log_action("CheckCredentials reply %s" % resp.strip())
        if resp.find("OK:") < 0:
            raise Exception("User authentication failed!")
        only_resp = (resp.upper().split("OK:")[-1]).rstrip("\n")
        if len(argin) == 3 and argin[2] == 'nofull':
            return only_resp.strip()
        UserId = only_resp.split(":")[0]
        propList = self.StorageMngrDev.ListProposal([UserId, self.TAG])
        invList = self.StorageMngrDev.ListInvestigation([UserId, self.TAG])
        return "\n".join([resp.strip(),propList.strip(),invList.strip()])


#------------------------------------------------------------------
#    Dump status into DB
#------------------------------------------------------------------
    def memorize_status(self):
        dict_in = {}
        dict_in['StaffMember'] = {'__value':self.StaffMember}
        dict_in['UserId'] = {'__value':self.UserId}
        dict_in['Logged_in'] = {'__value':self.Logged_in}
        dict_in['ActiveInvestigation'] = {'__value':self.ActiveInvestigation}
        self.db.put_device_attribute_property(self.get_name(),dict_in)

#------------------------------------------------------------------
#    Retrieve Proposal and Investigation List
#------------------------------------------------------------------
    def retrieve_proposal_investigations(self):
        if self.StorageMngrDev == None:
            self.connect_to_storage_manager()
        resp = self.StorageMngrDev.ListProposal([self.UserId, self.TAG])
        if resp.upper().find("OK:") >= 0:
            self.ProposalList = resp.split(":")
            self.ProposalList.pop(0)
        else:
            self.ProposalList = []
        for idx in range(len(self.ProposalList)):
            self.ProposalList[idx] = self.ProposalList[idx].strip()
        # Remove duplicates
        self.ProposalList = list(set(self.ProposalList))
        #
        resp = self.StorageMngrDev.ListInvestigation([self.UserId, self.TAG])
        if resp.upper().find("OK:") >= 0:
            self.InvestigationList = resp.split(":")
            self.InvestigationList.pop(0)
        else:
            self.InvestigationList = []
        for idx in range(len(self.InvestigationList)):
            self.InvestigationList[idx] = self.InvestigationList[idx].strip()
        # Remove duplicates and re-sort
        self.InvestigationList = sorted(list(set(self.InvestigationList)),key=str.lower)


#------------------------------------------------------------------
#    Set Active Investigation
#------------------------------------------------------------------
    def set_active_investigation(self):
        self.log_action("set_active_investigation %s" % self.ActiveInvestigation)
        self.retrieve_proposal_investigations()
        if self.ActiveInvestigation in self.InvestigationList:
            # Investigation already created
            res = 'OK'
            self.log_action("investigation already exists")
            return
        arg_in = [self.UserId, self.TAG, self.ActiveInvestigation, self.ActiveInvestigation]
        proposalFound = False
        for proposal in self.ProposalList:
            if self.ActiveInvestigation.startswith(proposal):
                arg_in[2] = proposal
                proposalFound = True
                self.log_action("found an existing proposal: %s" % proposal)
                break
        if not proposalFound and self.ActiveInvestigation[:8].isdigit():
            # VUO proposals with a corresponding Investigation are not listed in
            # self.ProposalList, let's suppose that the first part is a propsal ID
            proposal = self.ActiveInvestigation[:8]
            if self.StaffMember:
                # special case of a staff memeber
                arg_in[2] = proposal
                proposalFound = True
                self.log_action("staff member will try this proposal: %s" % proposal)
            else:
                # normal case of an external memeber
                for Investigation in self.InvestigationList:
                    if Investigation.startswith(proposal):
                        arg_in[2] = proposal
                        self.log_action("user will try this proposal: %s" % proposal)
                        proposalFound = True
                        break
        if proposalFound:
            res=self.StorageMngrDev.command_inout("CreateInvestigationFromProposal",arg_in)
            self.log_action("CreateInvestigationFromProposal() reply %s" % res)
            if res[:2] =='OK':
                return
        res = self.StorageMngrDev.command_inout("CreateInvestigation",arg_in)
        self.log_action("CreateInvestigation() reply %s" % res)
        if res[:2]!='OK':
            self.ActiveInvestigation = ""
            raise Exception("Create Investigation Failed!")

#------------------------------------------------------------------
#    Action internal log
#------------------------------------------------------------------
    def log_action(self, msg):
        self.ActionLog.append(msg)
        self.ActionLog = self.ActionLog[-100:]

#------------------------------------------------------------------
#    Storage manager connection
#------------------------------------------------------------------
    def connect_to_storage_manager(self):
        self.StorageMngrDev = PyTango.DeviceProxy(self.StorageManagerDevice)
        self.StorageMngrDev.set_timeout_millis(10000)

#------------------------------------------------------------------
#    Update disk usage info
#------------------------------------------------------------------
    def update_disk_usage_info(self):
         if (time.time() - self.free_space_readout_time) > 60:
             TB_FACTOR = float(1024 ** 3)
             self.free_space_readout_time = time.time()
             arg_in = [self.TAG]
             disk_usage = self.StorageMngrDev.DiskFree(arg_in)
             scr_avail = float(disk_usage.split(":")[0])
             scr_tot = float(disk_usage.split(":")[1])
             self.attr_FreeSpaceScratch_read = int(((scr_avail / scr_tot) * 100))
             self.attr_FreeSpaceScratch_TB_read = (scr_avail / TB_FACTOR)
             onl_avail = float(disk_usage.split(":")[2])
             onl_tot = float(disk_usage.split(":")[3])
             self.attr_FreeSpaceOnline_read = int(((onl_avail / onl_tot) * 100))
             self.attr_FreeSpaceOnline_TB_read = ( onl_avail / TB_FACTOR)

#==================================================================
#
#    Experiment_DetailsClass class definition
#
#==================================================================
class Experiment_DetailsClass(PyTango.DeviceClass):

    #    Class Properties
    class_property_list = {
        }


    #    Device Properties
    device_property_list = {
        'TAG':
            [PyTango.DevString,
            "TAG name of the experiment station",
            [ "LDM" ] ],
        'StorageManagerDevice':
            [PyTango.DevString,
            "Full Tango devicename of the storage manager",
            [ "storagemanager.esce.elettra.trieste.it:10000/StorageManager/storagemanager/1" ] ],
        }


    #    Command definitions
    cmd_list = {
        'CheckCredentials':
            [[PyTango.DevVarStringArray,"Username,Password"],
            [PyTango.DevString,""]],
        'Login':
            [[PyTango.DevString,"Password"],
            [PyTango.DevVoid,""]],
        'Logout':
            [[PyTango.DevVoid,""],
            [PyTango.DevVoid,""]],
            }


    #    Attribute definitions
    attr_list = {
        'UserLogin':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'Memorized':"true",
            } ],
        'ActiveInvestigation':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'Memorized':"false",
            } ],
        'Logged_in':
            [[PyTango.DevBoolean,
            PyTango.SCALAR,
            PyTango.READ]],
        'UserId':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ]],
        'StaffMember':
            [[PyTango.DevBoolean,
            PyTango.SCALAR,
            PyTango.READ]],
        'DataPath':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'Memorized':"false",
            } ],
        'ProposalList':
            [[PyTango.DevString,
            PyTango.SPECTRUM,
            PyTango.READ,10000]],
        'InvestigationList':
            [[PyTango.DevString,
            PyTango.SPECTRUM,
            PyTango.READ,10000]],
        'FreeSpaceScratch':
            [[PyTango.DevShort,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'unit':"%",
                'standard unit':"%",
                'display unit':"%",
            } ],
        'FreeSpaceOnline':
            [[PyTango.DevShort,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'unit':"%",
                'standard unit':"%",
                'display unit':"%",
            } ],
        'FreeSpaceScratch_TB':
            [[PyTango.DevFloat,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'unit':"TB",
                'standard unit':"TB",
                'display unit':"TB",
            } ],
        'FreeSpaceOnline_TB':
            [[PyTango.DevFloat,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'unit':"TB",
                'standard unit':"TB",
                'display unit':"TB",
            } ],
        'ActionLog':
            [[PyTango.DevString,
            PyTango.SPECTRUM,
            PyTango.READ,100]]
        }


#------------------------------------------------------------------
#    Experiment_DetailsClass Constructor
#------------------------------------------------------------------
    def __init__(self, name):
        PyTango.DeviceClass.__init__(self, name)
        self.set_type(name);

#==================================================================
#
#    Experiment_Details class main method
#
#==================================================================
def main():
    try:
        py = PyTango.Util(sys.argv)
        py.add_TgClass(Experiment_DetailsClass,Experiment_Details,'Experiment_Details')

        U = PyTango.Util.instance()
        U.server_init()
        U.server_run()

    except PyTango.DevFailed as e:
        print ('-------> Received a DevFailed exception:',e)
    except Exception as e:
        print ('-------> An unforeseen exception occured....',e)

if __name__ == '__main__':
    main()
